/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.modules.masterfs;

import java.io.IOException;
import java.io.File;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.WeakHashMap;
import org.netbeans.modules.masterfs.filebasedfs.fileobjects.WriteLock;
import org.netbeans.modules.masterfs.filebasedfs.fileobjects.WriteLockFactory;
import org.openide.filesystems.FileAlreadyLockedException;
import org.openide.filesystems.FileChangeListener;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.netbeans.modules.masterfs.filebasedfs.fileobjects.WriteLockUtils;
import org.openide.util.Exceptions;
import org.openide.util.WeakListeners;

/**
 * @author Radek Matous
 */
final class Delegate {
    private FileObject delegate;
    private FileObject secondDelegate;
    private FileChangeListener priorityWeakListener;
    private FileChangeListener versioningWeakListener;    
    final private FileChangeListener priorityListener;
    final private FileChangeListener versioningListener;
    private Reference lock = null;

    private int attribs = 0;

    static final int SECOND_DELEGATE = 1;
    static final int BEST_DELEGATE = 0;
    private final MasterFileObject mfo;

    Delegate(final FileObject fo, final FileChangeListener priorityListener, 
            final FileChangeListener versioningListener, MasterFileObject mfo) {
        this.priorityListener = priorityListener;
        this.versioningListener = versioningListener;
        this.mfo = mfo;                
    }

    FileObject get() {
        ResourcePath resPath = mfo.getResource();
        FileObject retVal = delegate;
        if (delegate instanceof InvalidDelegate) {
            retVal = null;       
        } else {
            if (retVal == null) {
                retVal = resolve(resPath, true);
                set(retVal);
            }
        }
        return retVal;
    }

    FileObject get(boolean maybeInvalid) {
        FileObject retVal = get ();
        if (maybeInvalid && retVal == null) {
            retVal = delegate;       
        } 
        return retVal;
    }
    
    static FileObject resolve(final ResourcePath resPath) {
        return resolve(resPath, true);
    }
    
    private static FileObject resolve(final ResourcePath resPath, final  boolean bestDelegate) {
        FileObject retVal;
        String normalizedPath = resPath.getNormalizedPath();
        if (WriteLockUtils.hasActiveLockFileSigns(normalizedPath)) {
            File f = new ResourcePath(normalizedPath).getFile();
            if (WriteLockUtils.hasActiveLockFileSigns(f.getName())) {
                return null;
            }
        }
        
        if (bestDelegate) {
            retVal = MountTable.getDefault().resolveBestDelegate(normalizedPath);
        } else {
            retVal = MountTable.getDefault().resolveSecondDelegate(normalizedPath);
        }

        if (retVal == null) {
            retVal = SpecialDelegates.get(resPath);
        } else {
            // reloads table from file .nbattrs. Changed values
            // may be cached, but previous delegate may change some values
            retVal.getAttribute("--hack--");//NOI18N
        }
        return retVal;
    }

    static FileLock getLockForDelegate(final FileLock testedLock, final FileObject delegate) throws IOException {
        if (testedLock == null)
            throw new IOException();

        if (!(testedLock instanceof FileLockImpl))
            throw new IOException();

        FileLockImpl fLockImpl = (FileLockImpl) testedLock;
        Delegate del = fLockImpl.getDelegate();

        if (del.getFileLockImpl() != fLockImpl) {
            throw new IOException();
        }

        FileLock retVal = fLockImpl.get(delegate);
        if (retVal == null)
            throw new IOException();

        return retVal;
    }
    
    FileObject getPrefered() {
        final FileObject retVal = get();
        FileObject secRetVal = null;
        if (retVal != null) {
            if (retVal.isRoot()) {
                secRetVal = getSecond();
            }
        }
        return (secRetVal != null) ? secRetVal : retVal;
    }

    void reset(final ResourcePath resPath) {
        FileObject lastDelegate = delegate;
        set(null);
        setSecond(null);
        if (get() == null) {
            boolean folder = (lastDelegate != null) ? lastDelegate.isFolder() : false;
            delegate = new InvalidDelegate(resPath, folder);
        } else {
            getSecond();
            attribs = 0;
            try {
                reLock();
            } catch (IOException e) {
                Exceptions.printStackTrace(e);
            }
        }
    }

    boolean isValid() {
        return (delegate != null) && (delegate.isValid());
    }

    public boolean isLocked() {
        FileLockImpl lck;
        String path = mfo.getPath ();
        synchronized (Delegate.class) {
            lck = getFileLockImpl();
        }
        return lck != null;
    }

    FileLock lock() throws IOException {
        FileLockImpl lck;
        String path = mfo.getPath ();

        synchronized (Delegate.class) {
            lck = getFileLockImpl();
            if (lck != null) throw new FileAlreadyLockedException(path);
            lock = new WeakReference(lck = new FileLockImpl());
        }

        lck.initLock(mfo.getResource().getFile(), get(), getPrefered());
        return lck;
    }

    boolean hasMountAbleFlag() {
        //TODO: evaluate commented code for Automount mode 
/*
        String path = mfo.getPath ();
        if ((attribs & MNTPOINTFLAG_SUPPRESS) == MNTPOINTFLAG_SUPPRESS) return false;
        if ((attribs & MNTPOINTFLAG) == MNTPOINTFLAG) return true;
        if ((attribs & MNTPOINTFLAG_RESOLVED) == MNTPOINTFLAG_RESOLVED) return false;

        attribs |= MNTPOINTFLAG_RESOLVED;


        if (ProviderCall.isMountAble(path)) {
            attribs |= MNTPOINTFLAG;
            return true;
        }
*/

        return false;
    }

    void set(final FileObject fo) {
        synchronized (this) {
            FileChangeListener oldPriorityListener = priorityWeakListener;
            FileChangeListener oldVersioningListener =  versioningWeakListener;
            FileObject oldDelegate = delegate;
            if (oldDelegate == fo) return;

            if (oldPriorityListener != null && oldDelegate != null)
                oldDelegate.removeFileChangeListener(oldPriorityListener);
                        
            if (oldVersioningListener != null && oldDelegate != null)
                oldDelegate.removeFileChangeListener(oldVersioningListener);

            delegate = fo;
            if (fo != null) {
                // the following is not typesafe, it is not nice,
                // and very likely it is not even correct, however until
                // there is a proper API for "priority listeners" in openide/fs
                // there is not much to do about this
                Class realClass;
                try {
                    realClass = Class.forName("org.openide.filesystems.FileObject$PriorityFileChangeListener");
                } catch (ClassNotFoundException ex) {
                    throw (IllegalStateException)new IllegalStateException(ex.getMessage()).initCause(ex);
                }
                priorityWeakListener = (FileChangeListener)WeakListeners.create(realClass, FileChangeListener.class, priorityListener, fo);
                fo.addFileChangeListener(priorityWeakListener);

                versioningWeakListener = (FileChangeListener)WeakListeners.create(FileChangeListener.class, FileChangeListener.class, versioningListener, fo);
                fo.addFileChangeListener(versioningWeakListener);                                 
            }
        }
    }

    private void reLock() throws IOException {
        final FileLockImpl lck = getFileLockImpl();
        if (lck != null) {
            lck.lock(delegate);
            lck.lock(secondDelegate);
        }
    }

    private FileLockImpl getFileLockImpl() {
        FileLockImpl lck;
        synchronized (Delegate.class) {
            lck = (FileLockImpl) ((lock == null) ? null : lock.get());
        }
        return lck;
    }

    private void setSecond(final FileObject fo) {
        synchronized (this) {
            secondDelegate = fo;
        }
    }

    private FileObject getSecond() {
        ResourcePath resPath = mfo.getResource();                
        FileObject retVal = secondDelegate;
        if (retVal == null) {
            retVal = resolve(resPath, false);
            setSecond(retVal);
        }
        return retVal;
    }

    private final class FileLockImpl extends FileLock {
        private final Map lockList = new WeakHashMap();
/*
        there is some ugly piece of code around mandatoryLock that ensures that also
        VCS support will use the same locking mechanism as FileBasedFileSystem does
*/
        private FileLock/**ever must be WriteLock*/ mandatoryLock;

        FileLockImpl() {
        }
        
        private void initLock(File file, final FileObject delegate, final FileObject prefered) throws IOException {
            try {
                lock(delegate);
                lock(prefered);
                if (mandatoryLock == null && file != null) {
                    if ((delegate != null && delegate.isData()) || (prefered != null && prefered.isData())) 
                    mandatoryLock = WriteLockFactory.tryLock(file, false);
                }
            } catch (IOException e) {
                releaseLock();
                throw e;                        
            } 
        }
        
        public void releaseLock() {
            if (this.isValid()) {
                super.releaseLock();                                      
                if (mandatoryLock != null) {
                    mandatoryLock.releaseLock(); 
                }
                final Collection locks = lockList.values();
                for (Iterator iterator = locks.iterator(); iterator.hasNext();) {
                    FileLock fLock = (FileLock) iterator.next();
                    fLock.releaseLock();
                }

                if (getCurrentLock () == this) {
                    synchronized (Delegate.class) {
                        lock = null;
                    }
                }
            }
            mfo.getExtensions().fileUnlocked(mfo);
        }

        private FileLock getCurrentLock () {
            return (lock != null) ? (FileLock)lock.get() : null;
        }
        
        private void lock(final FileObject delegate) throws IOException {
            if (delegate != null && this.get(delegate) == null) {
                FileLock fLock = delegate.lock();
                if (mandatoryLock == null && fLock instanceof WriteLock) {
                    mandatoryLock = fLock;
                }
                lockList.put(delegate, fLock);
            }
        }

        private FileLock get(final FileObject delegate) {
            return (FileLock) lockList.get(delegate);
        }

        private Delegate getDelegate() {
            return Delegate.this;
        }
    }
    private static class TestLockException extends IOException {
        Map m = new HashMap ();

        void put (FileLockImpl impl, String msg) {
           m.put(impl, new Exception(msg));
        }

        public void printStackTrace(FileLockImpl tested, FileLockImpl current) {
            Exception exc = (Exception)m.get(tested);
            if (exc != null) {
                System.err.println("-------- "+exc.getLocalizedMessage() + " ----------");
                exc.printStackTrace();
            }
            Exception exc2 = (Exception)m.get(current);
            if (exc2 != null) {
                System.err.println("-------- "+exc2.getLocalizedMessage() + " ----------");
                exc2.printStackTrace();
            }

        }
    }
    
    private static class InvalidDelegate extends InvalidDummy {
        private final boolean isFolder; 

        protected InvalidDelegate(ResourcePath resourcePath, boolean isFolder) {
            super(resourcePath);
            this.isFolder = isFolder;
        }

        public boolean isData() {
            return !isFolder;
        }

        public boolean isFolder() {
            return isFolder;
        }
    }
}
